#include "Lasers.h"

using namespace std;

VL53L1X sensors[nbrLasers];

///   I2C ADDRESS    ///

const uint8_t AddressI2CSensors[4] = {0x2D,0x2C,0x2B,0x2A};

///   PINS RESET SENSORS    ///
const uint8_t sensorsResetPins[4] = {5,18,23,32};

void initLaserSensors(void)
{
  pinMode(sensorsResetPins[0], OUTPUT);
  pinMode(sensorsResetPins[1], OUTPUT);
  pinMode(sensorsResetPins[2], OUTPUT);
  pinMode(sensorsResetPins[3], OUTPUT);

  digitalWrite(sensorsResetPins[0], LOW);
  digitalWrite(sensorsResetPins[1], LOW);
  digitalWrite(sensorsResetPins[2], LOW);
  digitalWrite(sensorsResetPins[3], LOW);

  Serial.println();
 
  for(int8_t i = 0; i<nbrLasers;i++)
  {
    digitalWrite(sensorsResetPins[i], HIGH);
    delay(100);
    sensors[i].setTimeout(500);
    if (!sensors[i].init())
    {
      Serial.print("Failed to detect and initialize sensor: ");Serial.println(i, DEC);
      while (1);
    }
    changeAddress(i, AddressI2CSensors[i]);
    sensors[i].startContinuous(50);
    Serial.print("VL53L1X ");Serial.print(i, DEC);Serial.println(" initialized");
    delay(100);
  }
}

void sensorRead(uint16_t* measure, uint8_t sensor_Id)
{
  *measure = sensors[sensor_Id].read();
  //if (sensor->timeoutOccurred()) { Serial.print(" TIMEOUT"); }
}

void changeAddress(uint8_t sensor_Id, uint8_t address)
{
  sensors[sensor_Id].setAddress(0x29);
  sensors[sensor_Id].setAddress(address);
}

void calculatePlane(uint16_t* measures, double* eq, uint8_t UpOrDown)
{
  const double matDown[3][3] = {(double)measures[UpOrDown]/10,(double)measures[LeftLaser]/10,(double)measures[RightLaser]/10,0,-1.675,1.675,-1.73,0,0};
  const double matUp[3][3] = {(double)measures[LeftLaser]/10,(double)measures[RightLaser]/10,(double)measures[UpOrDown]/10,-1.675,1.675,0,0,0,1.73};
  double matrix[3][3];
  if(UpOrDown)
    memcpy(matrix, matDown, sizeof(matDown));
  else
    memcpy(matrix, matUp, sizeof(matUp));
  
  double matrixT[3][3] = {0,0,0,0,0,0,0,0,0};
  double AT[3][3] = {0,0,0,0,0,0,0,0,0};
  
  for(int i = 0; i< 3; i++){
      for(int j = 0; j < 3; j++){
          matrixT[j][i] = matrix[i][j];
      }
  }
  
  double b[1][3];
  double A[3][3];
  
  for(int i = 0; i < 3; i++){
      b[0][i] = matrix[2][i];
  }
  
  for(int i = 0; i < 3; i++){
      for(int j = 0; j < 3; j++){
          A[j][i] = matrix[i][j];
      }
  }
  for(int i = 0; i < 3; i++)
      A[i][2] = 1;
  
  
  for(int i = 0; i < 3; i++){
      for(int j = 0; j < 3; j++){
         AT[j][i] = A[i][j];
      }
  }
  
  double ATA[3][3], ATAI[3][3], y[3][3], sum = 0;
  
  for (int i = 0; i < 3; i++){
      for (int j = 0; j < 3; j++){
          sum = 0;
          for(int k = 0; k < 3; k++)
              sum += AT[i][k]*A[k][j];
          ATA[i][j] = sum;
      }
  }
  
  y[0][0]=ATA[1][1]*ATA[2][2]-ATA[1][2]*ATA[2][1];        //***********************************
  y[0][1]=-1*(ATA[1][0]*ATA[2][2]-ATA[1][2]*ATA[2][0]);   //                                  *
  y[0][2]=ATA[1][0]*ATA[2][1]-ATA[1][1]*ATA[2][0];        //                                  *
  y[1][0]=-1*(ATA[0][1]*ATA[2][2]-ATA[2][1]*ATA[0][2]);   //                                  *
  y[1][1]=ATA[0][0]*ATA[2][2]-ATA[0][2]*ATA[2][0];        //    CALCULATION OF COFACTOR       *
  y[1][2]=-1*(ATA[0][0]*ATA[2][1]-ATA[0][1]*ATA[2][0]);   //                                  *
  y[2][0]=ATA[0][1]*ATA[1][2]-ATA[0][2]*ATA[1][1];        //                                  *
  y[2][1]=-1*(ATA[0][0]*ATA[1][2]-ATA[0][2]*ATA[1][0]);   //                                  *
  y[2][2]=ATA[0][0]*ATA[1][1]-ATA[0][1]*ATA[1][0];        //***********************************
  
  double det=ATA[0][0]*y[0][0]-ATA[0][1]*y[0][1]+ATA[0][2]*y[0][2];
  
  for(int i=0;i<3;i++)                        //For storing adjoint in another 2-D Array
  {
      for(int j=0;j<3;j++)
          ATAI[i][j]=y[j][i];
  }
  if(det!=0)
  {
      for(int i=0;i<3;i++)                       //For calculating inverse of matrix (2-D Array)
      {
          for(int j=0;j<3;j++)
              ATAI[i][j]/=det;
      }
  }
  else
    Serial.println("INVERSE NOT POSSIBLE AS DETERMINANT( |A| ) IS EQUAL TO ZERO");
  
  double ATAIAT[3][3];
  
  for (int i = 0; i < 3; i++){
      for (int j = 0; j < 3; j++){
          sum = 0;
          for(int k = 0; k < 3; k++)
              sum += ATAI[i][k]*AT[k][j];
          ATAIAT[i][j] = sum;
      }
  }
  
  double plane[1][3];
  
  for (int i = 0; i < 3; i++){
      for (int j = 0; j < 3; j++){
          sum = 0;
          for(int k = 0; k < 3; k++)
              sum += ATAIAT[i][k]*b[0][k];
          plane[0][i] = sum;
      }
  }
  for(uint8_t i = 0; i<3; i++)
    eq[i] = plane[0][i];
}

void tick_Lasers(uint16_t* measures, double* eqPlan1, double* eqPlan2)
{
  memset(measures, 0, 2*sizeof(measures));
  uint16_t sample;

  for(uint8_t j = 0;j<nbrSample;j++)
  {
    for(uint8_t i = 0;i<nbrLasers;i++)
    {
      sensorRead(&sample, i);
      measures[i]+=sample;
    }
  }

  for(uint8_t k = 0; k<nbrLasers;k++)
    measures[k]/=nbrSample;
  
  Serial.print("LaserUp : ");Serial.println(measures[0],DEC);
  Serial.print("LaserDown : ");Serial.println(measures[1],DEC);
  Serial.print("LaserRight : ");Serial.println(measures[2],DEC);
  Serial.print("LaserLeft : ");Serial.println(measures[3],DEC);
  
  calculatePlane(measures, eqPlan1, DownLaser);
  calculatePlane(measures, eqPlan2, UpLaser);

  Serial.print("eq0 : ");Serial.print(eqPlan1[0]);Serial.print("x + ");Serial.print(eqPlan1[1]);Serial.print("y + ");Serial.print(eqPlan1[2]);Serial.println("z");
  Serial.print("eq1 : ");Serial.print(eqPlan2[0]);Serial.print("x + ");Serial.print(eqPlan2[1]);Serial.print("y + ");Serial.print(eqPlan2[2]);Serial.println("z");
  
}
